
/* 
 * Name: conf.c
 *
 * Desc: routines to handle parsing HTTP requests out of a text file.
 *
 *  Copyright (C) 1999 Jason Moore
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

#include <config.h>

#include "conf.h"
#include "global.h"

/* Reads a \n terminated line from file descriptor (includes the \n) */
/* note: copied from Apache webserver project */

int readline (int fd, char *line) 
{
  int n, rc = 0;
  char c ;

  for (n = 1 ; n < BIG_BUFFER_SIZE ; n++)
    {
      if ((rc == read (fd, &c, 1)) != 1)
    {
      *line++ = c ;
      if (c == '\n')
        break ;
    }
      else if (rc == 0)
    {
      if (n == 1)
        return 0 ; /* EOF */
      else
        break ; /* unexpected EOF */
    }
      else 
    return -1 ;
    }

  *line = 0 ;
  return n ;
}


int build_request(char *url_string)
/*
 * given a url string, output an HTTP request for it on stdout
 *
 */
{
    struct url u;

    parse_url(url_string, &u); 

    /* create HTTP request on stdout */
    /* Note: Accept lines are taken from Netscape 4.5 Linux */

    printf("\
#
# %s HTTP request configuration file
#
# pipe this output to a file, edit it if required, and run %s again with -c
#
#  e.g.
#        %s -u http://www.sober.com/ > config.txt
#        %s -c config.txt
#
", APP_NAME, APP_NAME, APP_NAME, APP_NAME);
    
    printf("%s:%d\n", u.u_host, u.u_port);               // initial host:port line

    printf("GET %s%s HTTP/1.0\r\n", u.u_path, u.u_file);
    printf("Host: %s\r\n", u.u_host);
    printf("Accept: image/gif\r\n");
    printf("Accept: image/jpeg\r\n");
    printf("Accept: image/pjpeg\r\n");
    printf("Accept: image/png\r\n");
    printf("Accept: image/x-xbitmap\r\n");
    printf("Accept: */*\r\n");
    printf("Accept-Language: en\r\n");
    printf("User-Agent: %s %s %s \r\n\r\n", APP_NAME, VERSION, APP_URL);

    return(1);
}

/*
 *  Parse configuration file to create HTTP request
 *
 *  pass in an empty http request structure
 *  - not all fields in http_request struct are used 
 *      - struct url u
 *          - char u_service[24];
 *            char u_host[128];
 *            unsigned short u_port;
 *            char u_path[255];
 *            char u_file[128];
 *      - char req[] 
 */

int config_parse(char *filename, struct http_request *req)
{
    int fd;             // file descriptor
    char *p;            // pointer for parsing strings
    char *host;         // address/ip

    char line[256];     // line to read from config

    int count = 0;      // line number

    fd = open( filename, O_RDONLY );
    if (fd == -1)
    {
        fprintf(stderr, "Unable to read temp file: %s (%s)\n", filename, strerror(errno) );
        return(1);
    }

    #ifdef DEBUG
    fprintf(stderr, "config_parse: filename: %s\n", filename );
    #endif

    /* read data */

    while (readline(fd, line) != 0)
    {
            #ifdef DEBUG
                fprintf(stderr, "config_parse: line: %s", line);
            #endif 

            if (line[0] == '#')         // skip comments
                continue;

            if (count == 0)             // first non-comment line: read host:port
            {
                host = (char *)&line;

                p = strchr(line, ':');
                if (p != 0)
                {
                    char *end;
                    *p = 0;             // delimits host
                    p++;                // increase pointer past ':'
                    end = p;
                    /* do not include \r or \n in port */
                    while (*end != 0 && *end != '\r' && *end != '\n')
                    {
                        end++;
                    }
                    *end = 0;                    // mark end
                    req->u.u_port = atoi(p);
                }
                else
                {
                    req->u.u_port = DEFAULT_HTTP_PORT;
                }

                strcpy(req->u.u_host, host);            // copy host into struct
                strcpy(req->u.u_service, "http");       // set service

                #ifdef DEBUG
                fprintf(stderr, "HOST:[%s]\n", host);
                fprintf(stderr, "PORT:[%s]\n", p);
                #endif
            }
            else                        // add line to request
            {
                if (strcat(req->req, line) == 0)
                {
                    fprintf(stderr, "Unable to add line of request.\n");
                }
            }
            count++; 
    }
    close(fd);
    return(0);
}



#ifdef TEST


#include <sys/types.h>
#include <sys/stat.h>


int main(void)
{

    // char line[1024];
    char *tmpfile = "/tmp/wload.conf";
    struct http_request http_req;

    config_parse(tmpfile, &http_req);

    printf("HOST: %s\n", http_req.u.u_host);
    printf("PORT: %d\n", http_req.u.u_port); 
    printf("SERVICE: %s\n", http_req.u.u_service);
    printf("REQUEST:\n%s", http_req.req);

    exit(0);

}
#endif





