
/*
 * Name: http.c
 *
 * Desc: perform http requests
 *
 *  Copyright (C) 1999 Jason Moore
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <config.h>

#include "url.h"
#include "conf.h"
#include "display.h"
#include "http.h"
#include "msg.h"
#include "global.h"


/*
 *   http_request
 *
 *       - given an http_request struct
 *         open a socket and send the http request.
 *
 * TODO: use data in "http_request" to create socket
 *       - u.host u.port
 *
 *
 */

int http_request(struct http_request *r)
{
    int socket_fd;              // socket file-descriptor
    int    n;                      // line number for reads
    int total_r;                // total bytes read 

    char r_proto[9];            // response http protocol HTTP/1.0
    char r_code[4];             // response code 
    char r_msg[32];             // response messsage

    char recv_line[ MAXLINE + 1 ];        // buffer to read response
    struct timeval start_t, end_t;        // start and end timevals (sec/micro secs)
    struct sockaddr_in remote_sock;

    if (gettimeofday(&start_t, NULL) < 0)
        fprintf(stderr, "Error: Unable to gettimeofday.\n");

    r->start = start_t;                             // remember start time
    
    socket_fd = socket(AF_INET, SOCK_STREAM, 0);    // create socket
    bzero(&remote_sock, sizeof(remote_sock));       // init memory to zeros
    remote_sock.sin_family    = AF_INET;              // address family (unsigned short)
    remote_sock.sin_port    = htons(r->u.u_port);

    /* resolve host/ip */ 

    resolve_host(r);                                // addr is now the binary address

    remote_sock.sin_addr = r->u.u_addr;

    /* CONNECT */

    if ( connect(socket_fd, (struct sockaddr *) &remote_sock, sizeof(remote_sock) ) < 0)
        fprintf(stderr, "Unable to connect to socket: %s\n", strerror(errno) );
    
    /* SEND HTTP REQUEST */

    send(socket_fd, r->req, sizeof(r->req), 0);

    /* READ HTTP RESPONSE */

    // scan for HTTP status code (1xx, 2xx, 3xx, 4xx, 5xx)
    readline(socket_fd, recv_line);                           // first line
    sscanf(recv_line, "%s%s%s", r_proto, r_code, r_msg);      // e.g. HTTP/1.1 401 Not Found

    #ifdef DEBUG
    fprintf(stderr, "[response] line: %s", recv_line);      
    fprintf(stderr, "[response] proto: %s code: %s msg: %s\n", r_proto, r_code, r_msg);      
    #endif

    /* CHECK FOR ERROR (4xx, 5xx) */

    if (r_code[0] == '4' || r_code[0] == '5')
    {
        char status[64];
        if (r_code[0] == '4')
        {
            r->flag |= MSG_USER_ERROR;
            r->flag |= SAVE_RESPONSE;
        }
        else if (r_code[0] == '5')
        {
            r->flag |= MSG_SERVER_ERROR;
            r->flag |= SAVE_RESPONSE;
        }

        sprintf(status, "Warning: HTTP Error: %s %s (thread:%d)", r_code, r_msg, r->thread_id);
        display_status(status);

    }
    else    // reset flag (note: wipes out all settings)
    {
            r->flag = 0;
    }

    /* SAVE RESPONSE? */

    if (r->flag & SAVE_RESPONSE)
    {
        if (http_init_save_response(r) != 1)
            r->flag ^= SAVE_RESPONSE;                   // turn off saving - something is messed  
        else
            http_save_response(r, recv_line);           // save first line 
    }

    strcpy(r->status, r_code);

    /* READ REMAINDER OF RESPONSE */

    total_r = 0;
    while ( (n = read(socket_fd, recv_line, MAXLINE)) > 0)
    {
        recv_line[n] = 0;        // ??? (add null to end of string)
    
        total_r += n;

        if (r->flag & SAVE_RESPONSE)
            http_save_response(r, recv_line); 
    }

    #ifdef DEBUG
    fprintf(stderr, "r->flag: %d\n", r->flag);  
    #endif

    // close socket (or we'll run out...)
    close(socket_fd); 


    if (r->flag & SAVE_RESPONSE)
        http_end_save_response(r);

    /* SUMMARY */

    r->bytes = total_r;
    r->counter = (r->counter + 1);

    if (gettimeofday(&end_t, NULL) < 0)
        fprintf(stderr, "Unable to get end time: %d\n", errno);

    r->end = end_t;    
    
    return 0;
}

/*
 * sometimes you might want to:
 *     - force wload to use the same host for all threads RESOLVE_ONCE
 *     - let each thread lookup up the name independantly RESOLVE_THREAD
 *     - resolve the name for each request RESOLVE REQUEST (but since gethostbyname and others aren't thread safe, this
 *       really isn't viable.)
 *
 * currently, resolves host/ip to binary form once for each thread
 *
 * uses hostname from within the url struct in the http_request struct
 */

int resolve_host(struct http_request *r)
{

    struct hostent *host;

    if (r->u.u_host == '\0')
    {
        fprintf(stderr, "No host.\n");
        return(0);
    }

    // check if host has been resolved (only do it once)

    if (r->u.u_addr.s_addr != 0)
    {
        #ifdef DEBUG
        fprintf(stderr, "Host has been previously resolved. Skipping lookup.\n");
        #endif
        return(0);
    }

    host = gethostbyname(r->u.u_host);              // supports hosts and dotted-quads

    if (host == NULL)
    {
        fprintf(stderr, "Unable to gethostbyname: %s\n", strerror(h_errno) );
        return(0);
    }

    bcopy(host->h_addr_list[0], &r->u.u_addr.s_addr, host->h_length );      // save address

    #ifdef DEBUG
    fprintf(stderr, "resolve_host: host: %s IP: %s\n", r->u.u_host, inet_ntoa(r->u.u_addr) );
    #endif

    return(1);
}


/*
 * FILE* http_init_save_response(struct http_request *r);
 * http_save_response(FILE *f);
 * http_end_save_response
 *
 * - saves a response to a file (unique filename per thread)
 *
 */


int http_init_save_response(struct http_request *r)
{
    /* opens file for thread output */
    char filename[32];

    sprintf(filename, "wload-output-%02d", r->thread_id);  

    r->fh = fopen(filename, "w");
    if (r->fh == NULL)
    {
        fprintf(stderr, "Unable to open file: %s for writing: %s\n", filename, strerror(errno));
    }
    return(1);
}


int http_save_response(struct http_request *r, char *response)
{
    int b;
    b = strlen(response);
    if (fwrite(response, 1, b, r->fh) != b)
    {
        fprintf(stderr, "Unable to write full response to file (length:%d)\n", b); 
    }

    #ifdef DEBUG
    fprintf(stderr, "length of response: %d\n", b); 
    #endif

    return(1);
}

int http_end_save_response(struct http_request *r)
{
    fclose(r->fh);
    return(1);
}


#ifdef TEST

int main(void)
{
    // test request

    struct http_request r;

    strcpy(r.u.u_host, "www.sober.com");
    r.u.u_port = 81;

    strcpy(r.req, "GET / HTTP/1.0\n\n");

    http_request(&r);

    // printf("REQUEST: %s", r.req);

    return 1;
}

#endif





