
/*
 * Name: main.c
 *
 * Desc: parse command line opts, usage, thread creation
 *
 *  Copyright (C) 1999 Jason Moore
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


/*

    NOTES

    - (formerly "phread_net.c")
    - uses pthreads (glibc) 
    - socket code seems to support pthreads - install version of pthread
    which is included in glibc2 (since both libc's are installed, need to 
    pass special flags to gcc to use the correct includes, libs, compiler opts)
    (see: ~/orielly/examples/pthreads/glibc/Makefile )

    - threading model: uses pool of worker threads.

    alt) parent creates children threads.  child thread hits the server, reports back and 
    either gets sent out again, or killed.

    - uses a display thread to handle reading the message queue and updating the display
    - mostly static memory is used here.  hasn't been fine tuned.
    - reentrant socket stuff seems to be stable!

    - uses gnu options library (getopts) for command line options.
    - calcs average time


    TODO:

    - output raw dump of data which can be imported into spreadsheet
    - output file / thread for gnuplot

    - graphical ascii counter of hits e.g. ----, *---, **--

    - track slowest and fastest response time (for all threads)

    - how errno is handled with pthread functions (the global errno cannot be used.)
    - how to count how much data was read.. (simple technique currently)

    - improved error handling

    - should all children wait for others to complete.. or should they wait n seconds after each
    http request and try again.. (i would like to support both options) (it would be cool to see
    each box in the realtime report grid update independantly.  sort of a make 10 requests as fast as
    you can and then stop. 

    - warning if byte size differes more than n bytes
*/


#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <config.h>

#include "url.h"
#include "time_it.h"
#include "display.h"
#include "conf.h"
#include "global.h"


pthread_cond_t slave_driver         = PTHREAD_COND_INITIALIZER;    // crack the whip
pthread_cond_t new_message          = PTHREAD_COND_INITIALIZER;    // signal new message to the waiting message displayer 
pthread_mutex_t display_cond_lock   = PTHREAD_MUTEX_INITIALIZER;

static pthread_mutex_t request_cond_lock    = PTHREAD_MUTEX_INITIALIZER;    // used by worker/caller
static pthread_mutex_t worker_alive_lock    = PTHREAD_MUTEX_INITIALIZER;    // used to control access 
                                                                            // to the worker kill switch

static int alive = 1;                   // controls worker threads shutdown
int display_alive = 1;                  // controls display thread shutdown

int thread_id = 0;                      // each thread uses this number to get its id. 

struct timeval g_min, g_max;            // min/max for all threads

struct http_request init_req;           // used to seed request array - in order to initialize strings in struct
                                        // it needs to be here (not in main)

/* BEGIN */

int main(int argc, char **argv)
{
    int i;
    int num_threads             = DEFAULT_NUMBER_THREADS;   // number of threads
    int num_hits                = DEFAULT_NUMBER_REQUESTS;  // times to repeat (once for each thread)
    int delay_between_requests  = DEFAULT_DELAY;            // time delay (when worker threads are called en-masse)
    char ch;
    int get;                                                // for getch

    char request_file[64]       = "";                       // file of http request



    /* command line opts */

    while((ch = getopt(argc, argv, "hu:t:n:d:f:c:v")) != EOF) 
    {
        switch(ch) 
        {
              case 'h': { show_usage(); exit(1); }
              case 't': { num_threads = atoi(optarg); break; }            // number of threads
              case 'n': { num_hits = atoi(optarg); break; }               // number of requests
              case 'd': { delay_between_requests = atoi(optarg); break; }    // delay between thread calls
              case 'u': { build_request(optarg); exit(1); }
            case 'f': { strcpy(request_file, optarg); break; }    
            case 'c': { strcpy(request_file, optarg); break; }    
            default: { break; }    
        }
    }

    if (argc == 1 || (strcmp(request_file, "") == 0))    // exit if no args or no request file
    {
           show_usage();
           exit(1);
    }

    if (num_threads > MAX_THREADS)
    {
        num_threads = MAX_THREADS;
        fprintf(stderr, "Warning: number of threads set too high.  Using maximum: %d\n", MAX_THREADS);
    }


    /* create display thread */

    display_init();
    msg_init();

    if (pthread_create(&display_thread,
        NULL,
        (void *) display_update,
        (void *) NULL )            != 0)
            perror("pthread_create: cannot create display thread");



    /* MAIN PART of LOAD TEST */

    for (;;)
    {
        /* read config to fill HTTP request struct */

        /*  this initially loads the file into a global structure, but eventually, i would like
            the ability to have different requests for different threads.  (the ui to this would
            be a little tricky.)
        */ 

        #ifdef DEBUG
        fprintf(stderr, "<request type=empty>%s</request>\n", init_req.req);
        #endif

        config_parse(request_file, &init_req);

        #ifdef DEBUG
        fprintf(stderr, "<request>%s</request>\n", init_req.req);
        #endif

        for (i = 0; i < num_threads; i++)
        {
            /*    seed the request element structure */

            requests[i].thread_id    = i;                        // assign thread id
            requests[i].u.u_port    = init_req.u.u_port;        // assign port

            strcpy(requests[i].u.u_host, init_req.u.u_host);    // assign hostname
            strcpy(requests[i].req, init_req.req);              // assign http request

            if (pthread_create(&workers[i],
                NULL,
                (void *) worker_main,
                (void *) &requests[i])        != 0)
                    perror("pthread_create");
        }


        display_status("Ready to start load test.  Press a key to start");
        getch();                        // blocking getch
        nodelay(stdscr, TRUE);          // turn on non-blocking getch

        /* we now loop until finished - broadcast condition and wait */

        for (i = 0; i < num_hits; i++)
        {
            char header[80];
            sprintf(header, "Remaining requests: %d  Delay between requests: %d secs", (num_hits-i-1), delay_between_requests);
            display_status_header(header, 2);

            /* PAUSE */

            get = getch();
            if (get == ' ')             // spacebar pressed
            {
                display_status("** PAUSED **");
                nodelay(stdscr, FALSE);
                getch();
                display_status("");                     // clear pause msg
                nodelay(stdscr, TRUE);
            }

            if (i == (num_hits-1))                            // this is the last broadcast
            {
                #ifdef DEBUG
                fprintf(stderr, "at end. setting alive to zero.\n");
                #endif

                pthread_mutex_lock(&worker_alive_lock);
                alive = 0;
                pthread_mutex_unlock(&worker_alive_lock);

            }

            pthread_mutex_lock(&request_cond_lock);

                if (pthread_cond_broadcast(&slave_driver) != 0)            // broadcast wakes up all the threads
                {
                    perror("Unable to broadcast condition.\n"), exit(1);
                }

            pthread_mutex_unlock(&request_cond_lock);
            display_status("Sent broadcast message to start threads.");
            sleep(delay_between_requests);
        }

    
        /* wait for worker threads to finish */

        for (i = 0; i < num_threads; i++)
        {
            if (pthread_join(workers[i], NULL) != 0)
                perror("pthread_join"), exit(1);
        }

        /* wait for the display thread */
    
        display_alive = 0;
        pthread_cond_signal(&new_message);            // unlock sleeping display thread.

        if (pthread_join(display_thread, NULL) != 0)
            perror("pthread_join"), exit(1);
    
        display_status("All threads joined.");

        i = option_end();
        if (i == OPTION_EXIT)
            exit(0);

        if (i == OPTION_RESTART)
            continue;
    }
}



void worker_main(struct http_request *http)
/*
 * - wait for broadcast request
 * - do http request
 * - calculate totals 
 * - create message and send to display
 * - repeat...
 *
 */
{
    struct msg m;

    for (;;)
    {
        pthread_mutex_lock(&worker_alive_lock);    
        if (alive != 1) {
#ifdef DEBUG
            fprintf(stderr, "alive != 1 : exiting worker_main loop.\n");
#endif
            pthread_mutex_unlock(&worker_alive_lock);
            break;
        }
        pthread_mutex_unlock(&worker_alive_lock);

        pthread_mutex_lock(&request_cond_lock);
        pthread_cond_wait(&slave_driver, &request_cond_lock);
        pthread_mutex_unlock(&request_cond_lock);

        http_request(http);

        m.thread_id    = http->thread_id;
        m.counter    = http->counter;
        m.bytes        = http->bytes;
        m.info        = http->flag;
        strcpy(m.msg, http->status);

        time_diff(&http->start, &http->end, &http->dur);            // time to service request
        time_update_avg(&http->avg, http->counter, &http->dur);
        time_update_min(&http->min, &http->dur);
        time_update_max(&http->max, &http->dur);

        timestr(&http->dur, (char *)&m.t);
        timestr(&http->min, (char *)&m.t_min);
        timestr(&http->max, (char *)&m.t_max);
        timestr(&http->avg, (char *)&m.t_avg);
    
        msg_push(&m);                                // thread safe function

        #ifdef DEBUG
        fprintf(stderr, "About to send signal \"new_msg\"\n");
        #endif

        pthread_mutex_lock(&display_cond_lock);
        pthread_cond_signal(&new_message);
        pthread_mutex_unlock(&display_cond_lock);
    }
}


int option_end(void)
/*
 * interface for options for the end of the program.
 *
 */
{
    int i;

    display_status("Finished: [Q]uit");

    nodelay(stdscr, FALSE);                 // block

    for (;;)
    {
        i = tolower(getch());
        switch (i)
        {
            case 'q':
            {
                display_status("Quitting...");
                display_exit();
                return(OPTION_EXIT);
            }
            case 'r':
            {
                display_status("Restarting...(TODO)");
                return(OPTION_RESTART);
            }
            default:
            {
                display_status("Unknown option");
            }
        }
    }
}


void show_usage(void)
{
    printf("\
[%s v%s] - %s
Usage: %s [FLAGS] -f FILE
       %s -u URL

  -h    help - this message
  -t    number of threads             [%d]
  -n    number of requests            [%d]
  -d    delay between requests (secs) [%d]
  -r    raw output (turn off ncurses) (TODO)

  FILE  file containing HTTP request

  -u    parse a URL to createa an HTTP request file
        (output to stdout)

  homepage: %s

", APP_NAME, VERSION, APP_DESC, APP_NAME, APP_NAME, DEFAULT_NUMBER_THREADS, DEFAULT_NUMBER_REQUESTS, DEFAULT_DELAY, APP_URL);

}





