
/* msg.c */

/*

    - handles fetch/store of messages
    - reads queue to display

    - all msg_push msg_pop is thread safe

    Queue data structure

    - reader and writer indexes indicate what was last read/written

    - if reader == writer then the queue is full

    - if (reader + 1) == writer then queue is empty
        (note: if (reader+1) == MAX_QUEUE_LENGTH then if writer == 0, queue is empty)  

*/

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <pthread.h>

#include "msg.h"


struct msg_queue mq;                // global structure (public)

static pthread_mutex_t msg_queue_lock    = PTHREAD_MUTEX_INITIALIZER;        // protects global message queue
                                        // (private)


// 
// initialize message queue
//

int msg_init()
{
    mq.reader    = 0;                // reader has to "catch up" to writer            
    mq.writer    = 1;                // writer index points to the last message it wrote.

    #ifdef DEBUG
    fprintf(stderr, "Max queue size: %d\n", MAX_QUEUE_LENGTH );
    #endif

    return 0;
}


//
// "push" a message stucture onto the queue..
//

int msg_push(struct msg *m)    // msgs added by anyone (worker threads) 
{
    // returns: 0 on success
    //          -1 on full

    // note: this moves the "writer"

    struct msg *new_msg;

    pthread_mutex_lock(&msg_queue_lock);                /* LOCK */

    if ( mq.reader == mq.writer )
    {
        fprintf(stderr, "Message queue is full (%d).  Reader: %d Writer: %d\n", 
                    MAX_QUEUE_LENGTH, mq.reader, mq.writer);
        pthread_mutex_unlock(&msg_queue_lock);            /* UNLOCK */
        // i think a "goto" would be cleaner -- that way i know that the queue will 
        // get unlocked.. this is messy
        return(-1);
    }

    pthread_mutex_unlock(&msg_queue_lock);            /* UNLOCK */

    /* copy data from passing struct, to real array (queue) */
    
    new_msg = &mq.messages[mq.writer];

    new_msg->thread_id    = m->thread_id;
    new_msg->counter    = m->counter;
    new_msg->info       = m->info;

    #ifdef DEBUG
    //fprintf(stderr, "new_msg->thread_id:%d m->thread_id:%d\n", new_msg->thread_id, m->thread_id);
    #endif

    strcpy(new_msg->msg, m->msg);
    strcpy(new_msg->t, m->t);
    strcpy(new_msg->t_min, m->t_min);
    strcpy(new_msg->t_max, m->t_max);
    strcpy(new_msg->t_avg, m->t_avg);

    new_msg->bytes        = m->bytes;

    /* NOW increment the counter */

    pthread_mutex_lock(&msg_queue_lock);            /* LOCK */

    if ( (mq.writer+1) == MAX_QUEUE_LENGTH )    // writer is at physical end of array
    {
        mq.writer = 0;
    }
    else                // regular add
    {
        mq.writer = (mq.writer + 1);                // can i use ++ here? 
    }

    pthread_mutex_unlock(&msg_queue_lock);            /* UNLOCK */

    return 0;
}

// the queue is static, so a pointer to the message struct can be passed
// note: this is only called by the thread that is drawing to the screen (or file)
//     

/*
** non thread-safe functions
*/

int msg_unshift(struct msg **m)            
{
    // returns pointer to next message
    // returns -1 if no new messages to pop

    // note: this moves the "reader"

    //
    // note: only ONE thread will be calling this.. so it doesn't need to be locked..
    //       (this may be false economy... we'll see)
    //

    if ((mq.reader+1) == MAX_QUEUE_LENGTH)            // indexes are the same 
    {

        if (mq.writer == 0)
        {
            #ifdef DEBUG
            // fprintf(stderr, "msg_unshift: No new messages\n");
            #endif
            return -1;
        }
        mq.reader = 0;
    }
    else
    {
        if (mq.reader+1 == mq.writer)
        {
            #ifdef DEBUG
            // fprintf(stderr, "msg_unshift: No new messages\n");
            #endif
            return -1;
        }

        mq.reader = mq.reader + 1;
    }


    #ifdef DEBUG
    // fprintf(stderr, "msg_unshift: reading message from index:%d\n", mq.reader);
    #endif

    *m    = &mq.messages[mq.reader];

    return 0;
}







